%{
/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#define KITE_COMPILER
#include "kite_vm.h"
#include "objs/kite_object.h"

#define YYPARSE_PARAM parm
#define YYLEX_PARAM parm

/*
 * Dirty, dirty hack to make yyerror() work with Kite's exception system.
 * Credit: http://www.everything2.com/index.pl?node_id=1509675
 */
#define yyerror( msg )  dirty_yyerror( parm, msg )
void dirty_yyerror(void *, char *);
void checkReserved(void *, char *);

/* 
 * For some weird reason, we had to split the macro out this way
 * in order for the file/line numbering in exceptions to work properly.
 */
extern void *yyget_extra(void *);
extern int yylex(void *, void *, void *);

#define COMPILE_INSTRUCTION(i, j) \
    { \
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm); \
        kite_thread_t *thd = compiler->thd; \
        kite_object_t *cur = (kite_object_t*)compiler->obj_created->obj; \
        COMPILE_INSTRUCTION2(i, j, parm, &cur->builtin_data.funcvalue.funcptr); \
        (void)thd; \
        (void)cur; \
    }
    
void COMPILE_INSTRUCTION2(kite_opcode_t *inst, int lno, void *p, void **dest) 
    { 
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(p); 
        kite_object_t *cur = (kite_object_t*)compiler->obj_created->obj; 
        kite_opcode_t *fptr = *dest;
        inst->file = strdup(compiler->file); 
        inst->line = lno; 
        kite_add_to_instruction_list(&fptr, inst);
        *dest = fptr;
    }

%}

%token IMPORT
%token METHOD
%token SYMBOL_NAME
%token FROM
%token CLASS
%token CONSTRUCT
%token DESTRUCT
%token PROPERTY
%token GLOBAL
%token PLUS
%token MINUS
%token MULTIPLY
%token DIVIDE
%token UNARY_PLUS
%token UNARY_MINUS
%token MAP
%token REDUCE
%token ARRAY_DEREF
%token EQUALS
%token NOT_EQUALS
%token LESS_THAN
%token GREATER_THAN
%token LESS_OR_EQUALS
%token GREATER_OR_EQUALS
%token AND
%token OR
%token NOT
%token STRING_VALUE
%token FLOAT_VALUE
%token INT_VALUE
%token MOD
%token OPERATOR
%token XOR
%token LEFT_SHIFT
%token RIGHT_SHIFT
%token EVAL
%token DECIDE
%token BOOL_VALUE
%token UNTIL
%token WHILE
%token RUN
%token CATCH
%token IS
%token ISOF
%token MAKE
%token METHOD_CALL
%token NULL_VALUE

/* Needed since we can't quite lay out the associativity using the rules yet. */
%left MAP
%left REDUCE

%error-verbose
%defines
%pure-parser
%locations
%union {
    double doubleValue;
    long intValue;
    char *stringValue;
    struct kite_object_t *internalValue;
    kite_opcode_t *opValue;
}

%%

input:    /* empty */
    | statement ';' input
    | statement error ';' input
    ;

statement:
    | data_manipulation_statement
    | loop_statement
    | eval_statement
    | import_statement
    | class_definition
    | property_definition
    | decide_statement
    | run_catch_statement
    ;
    
import_statement:    IMPORT STRING_VALUE     {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_dereference_and_load(thd, $<stringValue>2);
        free($<stringValue>2);
    }
    ;
    
method_declaration: METHOD SYMBOL_NAME optional_param_name_list {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_push_stack(&compiler->obj_created,
                         kite_new_method(thd, FUNC_BYTECODE, NULL, 
                                        kite_list_count(thd, $<internalValue>3)));
        COMPILE_INSTRUCTION(kite_compile_funcargs($<internalValue>3), @1.first_line);
    } '[' input ']' {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *obj = kite_pop_stack(&compiler->obj_created);
        kite_object_t *parent = (kite_object_t*)compiler->obj_created->obj;
        kite_add_method(thd, parent, $<stringValue>2, obj);
        free($<stringValue>2);
    }
    ;

anon_method_declaration: METHOD optional_param_name_list     {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_push_stack(&compiler->obj_created,
                         kite_new_method(thd, FUNC_BYTECODE, NULL, 
                                        kite_list_count(thd, $<internalValue>2)));
        COMPILE_INSTRUCTION(kite_compile_funcargs($<internalValue>2), @1.first_line);
    } '[' input ']' {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_object_t *obj = kite_pop_stack(&compiler->obj_created);
        COMPILE_INSTRUCTION(kite_compile_push(obj), @4.first_line);
    }
    ;
        
optional_global:    {
        $<intValue>$ = 0;
    }
    | GLOBAL    {
        $<intValue>$ = 1;
    }
    ;
    
optional_param_name_list:    {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        $<internalValue>$ = kite_new_list(thd);
    }
    | param_name_list    {
        $<internalValue>$ = $<internalValue>1;
    }
    ;
    
param_name_list:    '(' ')'    {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        $<internalValue>$ = kite_new_list(thd);
    }
    | '(' param_elements ')'    {
        $<internalValue>$ = $<internalValue>2;
    }
    ;
    
param_elements: SYMBOL_NAME    {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        checkReserved(parm, $<stringValue>1);
        $<internalValue>$ = kite_new_list(thd);
        $<internalValue>$->builtin_data.listvalue.car = 
            kite_new_ident(thd, $<stringValue>1);
        free($<stringValue>1);
    }
    | SYMBOL_NAME ',' param_elements {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        checkReserved(parm, $<stringValue>1);
        $<internalValue>$ = kite_new_list(thd);
        $<internalValue>$->builtin_data.listvalue.car = 
            kite_new_ident(thd, $<stringValue>1);
        $<internalValue>$->builtin_data.listvalue.cdr = $<internalValue>3;
        free($<stringValue>1);
    }
    ;
    
class_definition:    CLASS SYMBOL_NAME {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_push_stack(&compiler->obj_created,
                         kite_new_class(thd, NULL, $<stringValue>2));
    } optional_from '[' class_body ']'    {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *obj = kite_pop_stack(&compiler->obj_created);
        kite_object_t *parent = (kite_object_t*)compiler->obj_created->obj;
        kite_add_property(thd, parent, $<stringValue>2, 1);
        kite_set_property(thd, parent, $<stringValue>2, obj);
        free($<stringValue>2);
    }
    ;
    
optional_from:        {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *obj = (kite_object_t*)compiler->obj_created->obj;
        obj->object_data.inherit_from = kite_dereference_and_load(thd, "System.object");
    }
    | FROM variable_immediate    {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *obj = (kite_object_t*)compiler->obj_created->obj;
        obj->object_data.inherit_from = $<internalValue>2;
        assert(obj->object_data.inherit_from->type == OBJ_CLASS ||
               obj->object_data.inherit_from->type == OBJ_INSTANCE);
    }
    ;

variable_immediate: SYMBOL_NAME {
        kite_funccall_info_t *fcall = malloc(sizeof(kite_funccall_info_t));
        void *opc = NULL;
        
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        COMPILE_INSTRUCTION2(kite_compile_push(kite_new_ident(thd, $<stringValue>1)), 
                             @1.first_line, parm, &opc);
        COMPILE_INSTRUCTION2(kite_compile_deref_1(TRUE), @1.first_line, parm, &opc);
        
        fcall->file = compiler->file;
        fcall->line = @1.first_line;
        fcall->locals = NULL;
        fcall->this = (kite_object_t*)compiler->obj_created->next->obj;
        kite_push_stack(&thd->func_stack, fcall);
        
        kite_vm_execute_user_method(thd, opc);
        
        kite_pop_stack(&thd->func_stack);
        free(fcall);
        
        $<internalValue>$ = kite_vm_pop(thd);
        free($<stringValue>1);
        kite_free_instruction_list(thd, opc);
    }
    | variable_immediate '.' SYMBOL_NAME {
        void *opc = NULL;
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        COMPILE_INSTRUCTION2(kite_compile_push($<internalValue>1), 
                             @2.first_line, parm, &opc);
        COMPILE_INSTRUCTION2(kite_compile_push(kite_new_ident(thd, $<stringValue>3)), 
                             @2.first_line, parm, &opc);
        COMPILE_INSTRUCTION2(kite_compile_deref_2(), @2.first_line, parm, &opc);
        kite_vm_execute_user_method(thd, opc);
        $<internalValue>$ = kite_vm_pop(thd);
        free($<stringValue>3);
        kite_free_instruction_list(thd, opc);
    }
    ;
    
class_body:    
    | class_elements
    ;

class_elements:    class_element
    | class_elements ',' class_element
    ;
    
class_element:    constructor_definition
    | destructor_definition
    | property_definition
    | method_declaration
    | operator_definition
    ;

constructor_definition:    CONSTRUCT optional_param_name_list {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_push_stack(&compiler->obj_created,
                         kite_new_method(thd, FUNC_BYTECODE, NULL, 
                                        kite_list_count(thd, $<internalValue>2)));
        COMPILE_INSTRUCTION(kite_compile_funcargs($<internalValue>2), @1.first_line);
    } '[' input ']' {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *obj = kite_pop_stack(&compiler->obj_created);
        kite_object_t *parent = (kite_object_t*)compiler->obj_created->obj;
        kite_add_method(thd, parent, "__construct__", obj);
    }
    ;
    
destructor_definition: DESTRUCT {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_push_stack(&compiler->obj_created,
                         kite_new_method(thd, FUNC_BYTECODE, NULL, 0));
    } '[' input ']' {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_object_t *obj = kite_pop_stack(&compiler->obj_created);
        kite_object_t *parent = (kite_object_t*)compiler->obj_created->obj;
        parent->object_data.destructor = obj;
    }
    ;
    
property_definition:    optional_global PROPERTY SYMBOL_NAME    {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *parent = (kite_object_t*)compiler->obj_created->obj;
        
        if (parent->type == OBJ_METHOD) {
            COMPILE_INSTRUCTION(
                kite_compile_push(kite_new_ident(thd, $<stringValue>3)), @2.first_line);
            COMPILE_INSTRUCTION(kite_compile_defprop(), @2.first_line);
        } else {
            kite_add_property(thd, parent, $<stringValue>3, $<intValue>1);
        }
        free($<stringValue>3);
    }
    ;

operator_definition:    OPERATOR operator_type param_name_list {
    kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
    kite_thread_t *thd = compiler->thd;
    kite_push_stack(&compiler->obj_created,
                     kite_new_method(thd, FUNC_BYTECODE, NULL, 
                                    kite_list_count(thd, $<internalValue>3)));
    COMPILE_INSTRUCTION(kite_compile_funcargs($<internalValue>3), @1.first_line);
} '[' input ']' {
    kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
    kite_thread_t *thd = compiler->thd;
    kite_object_t *obj = kite_pop_stack(&compiler->obj_created);
    kite_object_t *parent = (kite_object_t*)compiler->obj_created->obj;
    kite_add_operator(thd, parent, $<intValue>2, obj);
}
    ;
    
operator_type:    PLUS    {
        $<intValue>$ = OP_ADD;
    }
    | MINUS    {
        $<intValue>$ = OP_SUBTRACT;
    }
    | MULTIPLY     {
        $<intValue>$ = OP_MULTIPLY;
    }
    | DIVIDE    {
        $<intValue>$ = OP_DIVIDE;
    }
    | MOD    {
        $<intValue>$ = OP_MODULUS;
    }
    | UNARY_PLUS    {
        $<intValue>$ = OP_UNARY_PLUS;
    }
    | UNARY_MINUS    {
        $<intValue>$ = OP_UNARY_MINUS;
    }
    | MAP    {
        $<intValue>$ = OP_MAP;
    }
    | REDUCE    {
        $<intValue>$ = OP_REDUCE;
    }
    | ARRAY_DEREF    {
        $<intValue>$ = OP_ARRAY_DEREF;
    }
    | EQUALS    {
        $<intValue>$ = OP_EQUALS;
    }
    | NOT_EQUALS    {
        $<intValue>$ = OP_NOT_EQUALS;
    }
    | LESS_THAN    {
        $<intValue>$ = OP_LESS_THAN;
    }
    | GREATER_THAN    {
        $<intValue>$ = OP_GREATER_THAN;
    }
    | LESS_OR_EQUALS    {
        $<intValue>$ = OP_LESS_OR_EQUALS;
    }
    | GREATER_OR_EQUALS    {
        $<intValue>$ = OP_GREATER_OR_EQUALS;
    }
    | AND    {
        $<intValue>$ = OP_AND;
    }
    | OR    {
        $<intValue>$ = OP_OR;
    }
    | NOT    {
        $<intValue>$ = OP_NOT;
    }
    | XOR    {
        $<intValue>$ = OP_XOR;
    }
    | LEFT_SHIFT    {
        $<intValue>$ = OP_LEFT_SHIFT;
    }
    | RIGHT_SHIFT    {
        $<intValue>$ = OP_RIGHT_SHIFT;
    }
    | METHOD_CALL    {
        $<intValue>$ = OP_METHOD_CALL;
    }
    ;

data_manipulation_statement:    assignment_statement
    | method_declaration
    ;
    
assignment_statement:    var_name '=' data_manipulation_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_ASSIGNMENT), @2.first_line);
    }
    | or_statement
    ;
    
or_statement:        or_statement {
        COMPILE_INSTRUCTION(kite_compile_dupe_top(), @1.first_line);
        $<opValue>$ = kite_compile_jump_true(NULL);
        COMPILE_INSTRUCTION($<opValue>$, @1.first_line);
    } OR xor_statement {
        kite_opcode_t *opc = kite_compile_nop();
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_OR), @3.first_line);
        ((kite_opcode_jump_true*)$<opValue>2)->jumpto = opc;
        COMPILE_INSTRUCTION(opc, @3.first_line);
    }
    | xor_statement
    ;

xor_statement:    xor_statement XOR and_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_XOR), @2.first_line);
    }
    | and_statement
    ;
    
and_statement:    and_statement {
        COMPILE_INSTRUCTION(kite_compile_dupe_top(), @1.first_line);
        $<opValue>$ = kite_compile_jump_false(NULL);
        COMPILE_INSTRUCTION($<opValue>$, @1.first_line);
    } AND bitwise_statement {
        kite_opcode_t *opc = kite_compile_nop();
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_AND), @3.first_line);
        ((kite_opcode_jump_true*)$<opValue>2)->jumpto = opc;
        COMPILE_INSTRUCTION(opc, @3.first_line);
    }
    | bitwise_statement
    ;

bitwise_statement:    bitwise_statement LEFT_SHIFT equality_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_LEFT_SHIFT), @2.first_line);
    }
    | bitwise_statement RIGHT_SHIFT equality_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_RIGHT_SHIFT), @2.first_line);
    }
    | equality_statement
    ;
    
equality_statement:    equality_statement EQUALS less_greater_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_EQUALS), @2.first_line);
    }
    | equality_statement NOT_EQUALS less_greater_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_NOT_EQUALS), @2.first_line);
    }
    | equality_statement IS less_greater_statement {
        COMPILE_INSTRUCTION(kite_compile_objis(FALSE), @2.first_line);
    }
    | equality_statement ISOF less_greater_statement {
        COMPILE_INSTRUCTION(kite_compile_objis(TRUE), @2.first_line);
    }
    | less_greater_statement
    ;
    
less_greater_statement:    less_greater_statement '<' add_sub_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_LESS_THAN), @2.first_line);
    }
    | less_greater_statement '>' add_sub_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_GREATER_THAN), @2.first_line);
    }
    | less_greater_statement LESS_OR_EQUALS add_sub_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_LESS_OR_EQUALS), @2.first_line);
    }
    | less_greater_statement GREATER_OR_EQUALS add_sub_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_GREATER_OR_EQUALS), @2.first_line);
    }
    | add_sub_statement
    ;
    
add_sub_statement:    add_sub_statement '+' multdiv_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_ADD), @2.first_line);
    }
    | add_sub_statement '-' multdiv_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_SUBTRACT), @2.first_line);
    }
    | multdiv_statement
    ;
    
multdiv_statement:    multdiv_statement '*' unary_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_MULTIPLY), @2.first_line);
    }
    | multdiv_statement '/' unary_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_DIVIDE), @2.first_line);
    }
    | multdiv_statement '%' unary_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_MODULUS), @2.first_line);
    }
    | unary_statement
    ;
    
unary_statement: '+' unary_statement {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_null(thd)), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_UNARY_PLUS), @1.first_line);
    }
    | '-' unary_statement {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_null(thd)), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_UNARY_MINUS), @1.first_line);
    }
    | NOT unary_statement {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_null(thd)), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_NOT), @1.first_line);
    }
    | map_reduce_statement
    ;
    
map_reduce_statement:    deref_statement MAP deref_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_MAP), @2.first_line);
    }
    | deref_statement REDUCE deref_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_REDUCE), @2.first_line);
    }
    | deref_statement
    ;
    
deref_statement:    deref_statement {
        /*COMPILE_INSTRUCTION(kite_compile_dupe_top(), @1.first_line);*/
    } '|' SYMBOL_NAME {
        char *str = $<stringValue>4;
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_ident(thd, str)), @3.first_line);
        free(str);
    } optional_param_list {
        COMPILE_INSTRUCTION(kite_compile_call(), @3.first_line);
    }
    | deref_statement '[' data_manipulation_statement ']'    {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_ARRAY_DEREF), @2.first_line);
    }
    | SYMBOL_NAME {
        char *str = $<stringValue>1;
		COMPILE_INSTRUCTION(kite_compile_this(), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_ident(thd, str)), @1.first_line);
        free(str);
        COMPILE_INSTRUCTION(kite_compile_deref_1(FALSE), @1.first_line);
    } param_list {
        COMPILE_INSTRUCTION(kite_compile_call(), @1.first_line);
    }
    | constant_statement
    ;

optional_param_list: {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_list(thd)), 0);
    }
    | param_list
    ;

param_list:    '(' ')' {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_list(thd)), @1.first_line);
    }
    | '(' params_list ')'
    ;

params_list: data_manipulation_statement    {
        COMPILE_INSTRUCTION(kite_compile_list_cons_1(thd), @1.first_line);
    }
    | data_manipulation_statement ',' params_list    {
        COMPILE_INSTRUCTION(kite_compile_list_cons(thd), @2.first_line);
    }
    ;
    
var_name:    SYMBOL_NAME {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_ident(thd, $<stringValue>1)),
             @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_deref_1(TRUE), @1.first_line);
        free($<stringValue>1);
    }
    | var_name '.' SYMBOL_NAME {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_ident(thd, $<stringValue>3)),
             @2.first_line);
        COMPILE_INSTRUCTION(kite_compile_deref_2(), @2.first_line);
        free($<stringValue>3);
    }
    ;
    
constant_statement:    STRING_VALUE     {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_string(thd, $<stringValue>1)), 
            @1.first_line);
        free($<stringValue>1);
    }
    | INT_VALUE {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_integer(thd, $<intValue>1)), 
            @1.first_line);
    }
    | FLOAT_VALUE {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_float(thd, $<doubleValue>1)), 
            @1.first_line);
    }
    | BOOL_VALUE {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_boolean(thd, $<intValue>1)), 
            @1.first_line);
    }
    | NULL_VALUE {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_null(thd)), @1.first_line);
    }
    | var_name
    | list_value
    | construct_object_statement
    | anon_method_declaration
    | '(' data_manipulation_statement ')'
    ;

construct_object_statement:    MAKE var_name param_list    {
        COMPILE_INSTRUCTION(kite_compile_make(), @1.first_line);
    }
    ;
    
list_value: '[' optional_list_elements ']'
    ;
    
optional_list_elements:    {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_list(thd)), 0);
    }
    | list_elements
    ;
    
list_elements: list_element    {
        COMPILE_INSTRUCTION(kite_compile_list_cons_1(thd), @1.first_line);
    }
    | list_element ',' list_elements    {
        COMPILE_INSTRUCTION(kite_compile_list_cons(thd), @2.first_line);
    }
    ;

list_element:    or_statement
    ;
    
decide_statement:    DECIDE {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_push_stack(&compiler->decide_exits, kite_compile_nop());
    } '[' decide_contents ']' {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_opcode_t *tmpopc = kite_pop_stack(&compiler->decide_exits);
        COMPILE_INSTRUCTION(tmpopc, @5.first_line);
    }
    ;
    
decide_contents:    decide_condition
    | decide_contents ',' decide_condition
    ;
    
decide_condition:    constant_statement {
        $<opValue>$ = kite_compile_jump_true(NULL);
        COMPILE_INSTRUCTION($<opValue>$, @1.first_line);
    } '[' input ']'    {
        kite_opcode_jump_true *c = (kite_opcode_jump_true*)$<opValue>2;
        c->jumpto = c->common.next;
        c->common.next = kite_compile_nop();
        kite_opcode_t *tmp = c->jumpto;
        while(tmp->next != NULL) {
            tmp = tmp->next;
        }
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        tmp->next = compiler->decide_exits->obj;
    }
    ;
    
eval_statement:    EVAL {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *c = kite_dereference_and_load(thd, "System.vm.compiler");
        COMPILE_INSTRUCTION(kite_compile_push(c), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_ident(thd, "compileString")), 
            @1.first_line);
    } or_statement    {
        COMPILE_INSTRUCTION(kite_compile_list_cons_1(thd), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_call(), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_eval(), @1.first_line);
    }
    ;

loop_statement:    loop_type '[' input ']'    {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_boolean(thd, TRUE)), @2.first_line);
        COMPILE_INSTRUCTION(kite_compile_jump_true($<opValue>1), @2.first_line);
        
        /* Fix jump destination */
        kite_opcode_t *opc = $<opValue>1;
tryagain:
        while(opc->opcode != JUMP_TRUE && opc->opcode != JUMP_FALSE) {
            opc = opc->next;
        }
        if ((opc->opcode == JUMP_TRUE && ((kite_opcode_jump_true*)opc)->jumpto != NULL) ||
            (opc->opcode == JUMP_FALSE && ((kite_opcode_jump_false*)opc)->jumpto != NULL)) {
            opc = opc->next;
            goto tryagain;
        }
        ((kite_opcode_jump_true*)opc)->jumpto = kite_compile_nop();
        COMPILE_INSTRUCTION(((kite_opcode_jump_true*)opc)->jumpto, @2.first_line);
    }
    ;

loop_type:    WHILE {
        $<opValue>$ = kite_compile_nop();
        COMPILE_INSTRUCTION($<opValue>$, @1.first_line);
    } constant_statement    {
        COMPILE_INSTRUCTION(kite_compile_jump_false(NULL), @1.first_line);
        $<opValue>$ = $<opValue>2;
    }
    | UNTIL {
        $<opValue>$ = kite_compile_nop();
        COMPILE_INSTRUCTION($<opValue>$, @1.first_line);
    } constant_statement    {
        COMPILE_INSTRUCTION(kite_compile_jump_true(NULL), @1.first_line);
        $<opValue>$ = $<opValue>2;
    }
    ;

run_catch_statement: RUN {
        $<opValue>$ = kite_compile_exception_hndl(NULL);
        COMPILE_INSTRUCTION($<opValue>$, @1.first_line);
    } '[' input ']' {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_boolean(thd, TRUE)), @1.first_line);
        $<opValue>$ = kite_compile_jump_true(NULL);
        COMPILE_INSTRUCTION($<opValue>$, @1.first_line);
    } CATCH {
        kite_opcode_t *opc = kite_compile_nop();
        ((kite_opcode_exception_hndl*)$<opValue>2)->handler = opc;
        COMPILE_INSTRUCTION(opc, @7.first_line);
    } '[' input ']' {
        kite_opcode_t *opc = kite_compile_pop_exception_hndl();
        kite_opcode_t *nop = kite_compile_nop();
        ((kite_opcode_jump_true*)$<opValue>6)->jumpto = opc;
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_boolean(thd, TRUE)), @7.first_line);
        COMPILE_INSTRUCTION(kite_compile_jump_true(nop), @7.first_line);
        COMPILE_INSTRUCTION(opc, @7.first_line);
        COMPILE_INSTRUCTION(nop, @7.first_line);
    }
    ;
    
%%

void dirty_yyerror(void *yyparse_param, char *str) {
    kite_object_t *exc;
    kite_compiler_t *comp = (kite_compiler_t*)yyget_extra(yyparse_param);
    kite_thread_t *thd = comp->thd;
    
    kite_funccall_info_t *entry = calloc(sizeof(kite_funccall_info_t), 1);
    entry->this = NULL;
    entry->file = comp->file;
    entry->line = comp->currentLine;
    kite_push_stack(&thd->func_stack, entry);

    exc = kite_new_exception(thd, "System.exceptions.SyntaxError", 
                             str);
    kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);    
}

void checkReserved(void *p, char *name) {
    if (!strcmp(name, "__exc") ||
        !strcmp(name, "this")) {
        dirty_yyerror(p, "Cannot use reserved names for parameters.");
    }
}
