/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <assert.h>
#include "kite_vm.h"
#include "objs/kite_object.h"

/*****************************************************************************
 * Create a new instance of the VM
 ****************************************************************************/
kite_vm_t *kite_new_vm(char **args)
{
    kite_vm_t *ret;
    kite_thread_t *thd;
    int i;

    ret = calloc(1, sizeof(kite_vm_t));
    if (!ret) return NULL;
    
    ret->numthreads = 0;
    ret->thread_list = NULL;
    thd = kite_new_thread_without_start(ret);
    ret->root_package = kite_new_class(thd, NULL, "");
    ret->args = args;
    ret->gc_begin = ret->gc_end = NULL;
    pthread_mutex_init(&ret->gc_mtx, NULL);

    /* Clear signals */
    for(i = 0; i < NSIG; i++) {
        signal(i, SIG_DFL);
    }

    return ret;
}

/*****************************************************************************
 * Destroys a VM
 ****************************************************************************/
void kite_free_vm(kite_vm_t **vm)
{
    kite_vm_t *cur;
    int i;
    assert(vm != NULL && (*vm) != NULL);
    
    /* Clear signals */
    for(i = 0; i < NSIG; i++) {
        signal(i, SIG_DFL);
    }

    /* Kill all threads */
    cur = *vm;
    for ( ; cur->numthreads > 0; cur->numthreads--)
    {
        kite_exit_thread(cur, cur->thread_list[cur->numthreads - 1]);
    }

    /* Begin garbage collection until everything's collected */
    kite_thread_t *thd = kite_new_thread_without_start(*vm);
    kite_gc_destroy_all(thd);
    kite_destruct_object(thd, &(*vm)->root_package, FALSE);
    kite_exit_thread(cur, thd);

    pthread_mutex_destroy(&cur->gc_mtx);
    free(cur);
    *vm = NULL;
}

